package ca.recrown.islandsurvivalcraft.utilities.pathfinding;

import java.util.HashSet;
import java.util.PriorityQueue;
import java.util.Queue;

import ca.recrown.islandsurvivalcraft.utilities.datatypes.Point2;

public class DepthFirstSearch {
    private final HashSet<Node> checkedNodes;
    private final NodeComparable comparable;
    private final Queue<Node> queue;
    private final Node startNode;
    private final Point2 endGoal;
    private final int maxNodesSearched; 
    private Node currentValidNode;
    private Node results;

    /**
     * Instantiates a DFS search object.
     * @param start The starting point for this search.
     * @param endGoal The objective. May be null, but a target validator will then be required.
     * @param customComparable The comparable to use to compare points. May be null, where then there will be no prioritization of direction.
     * @param maxNodesSearched The maximum amount of allowed searched nodes. set to 0 for infinite.
     */
    public DepthFirstSearch(Point2 start, Point2 endGoal, NodeComparable customComparable, int maxNodesSearched) {
        if (start == null) throw new NullPointerException("Start point cannot be null.");
        checkedNodes = new HashSet<>();
        queue = new PriorityQueue<>();
        this.comparable = customComparable;
        this.endGoal = endGoal;
        this.startNode = new Node(null, start, endGoal, customComparable);
        queue.add(startNode);
        this.maxNodesSearched = maxNodesSearched;
    }

    /**
     * Instantiates a DFS search object.
     * Where there is no comparable set and distance to goal will be used if there is one,
     * and where there is no limit on amount of searched nodes.
     * @param start The starting point of the search.
     * @param endGoal The objective of the search, may be null.
     */
    public DepthFirstSearch(Point2 start, Point2 endGoal) {
        this(start, endGoal, null, 0);
    }


    public Node getResult() {
        return results;
    }

    /**
     * @return the currentNode
     */
    public Node getCurrentNode() {
        return currentValidNode;
    }
    
    /**
     * Makes a search step.
     * @param validity the validator.
     * @param objective used to determine whether or not this DFS has founds its objective. Will use this over using the end goal.
     * @return true if done, whether or not it found the objective is determined by the result.
     */
    public boolean step(CoordinateValidatable validity, CoordinateValidatable objective) {
        if (validity == null) throw new NullPointerException("The validator cannot be null.");
        Node node = queue.poll();
        if (node == null) return true;

        if (checkedNodes.add(node) && validity.validate(node)) {
            currentValidNode = node;
            if (objective != null) {
                if (objective.validate(currentValidNode)) {
                    results = currentValidNode;
                    return true;
                }
            } else {
                if (endGoal == null) throw new IllegalStateException("Cannot build path without a target or goal.");
                if (endGoal.fastEquals(currentValidNode)) {
                    results = currentValidNode;
                    return true;
                }
            }
            if (maxNodesSearched > 0 && getSearchedCount() >= maxNodesSearched) return true;
            Node north = new Node(currentValidNode, currentValidNode.x, currentValidNode.y + 1, endGoal, comparable);
            queue.add(north);
            Node south = new Node(currentValidNode, currentValidNode.x, currentValidNode.y - 1, endGoal, comparable);
            queue.add(south);
            Node west = new Node(currentValidNode, currentValidNode.x - 1, currentValidNode.y, endGoal, comparable);
            queue.add(west);
            Node east = new Node(currentValidNode, currentValidNode.x + 1, currentValidNode.y, endGoal, comparable);
            queue.add(east);
        }
        return false;
    }

    /**
     * Searches until complete. Completion is defined such that when a call to the step function does not return true.
     * @param validator The validator to use to check if a node is considered allowed to be used.
     * @param targetValidatable The validator to be used to determine whether or not the current node is the objective. May be null, however, the end goal must not be null in such a case.
     * @return
     */
    public boolean buildToGoal(CoordinateValidatable validator, CoordinateValidatable targetValidatable) {
        while (!step(validator, targetValidatable));
        return results != null;
    }

    public int getSearchedCount() {
        return checkedNodes.size();
    }
}